﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace TinyBravers {

	public class TinyPlayerController : TinyController {

		Camera mainCamera;

		protected override void AwakeInit() {
			base.AwakeInit();

			mainCamera = Camera.main;
		}

		protected override void StartInit() {
			base.StartInit();

			// Since I put this script in the same prefab as the character we can just possess it directly
			Possess(GetComponent<TinyPlayableCharacter>());

			// Start the pathfinding hover search
			StartCoroutine(PathFindingMovementHover());
		}

		/// <summary>
		/// This function raycast from the mouse and shows the green path finding markers on the floor.
		/// </summary>
		/// <returns></returns>
		IEnumerator PathFindingMovementHover() {
			while (true) {
				if (_character != null && _character.CurrentState == CharacterState.Waiting) {
					RaycastHit hit;
					Ray ray = mainCamera.ScreenPointToRay(Input.mousePosition);

					if (Physics.Raycast(ray, out hit)) {
						TinyTile tileHit = hit.transform != null ? hit.transform.parent != null ? hit.transform.parent.GetComponent<TinyTile>() : null : null;

						if (tileHit != null) {
							TinyTileMapManager.instance.ShowPathFindingMarkers(_character.transform.position, hit.point);
						}
					}
				}

				// Here is how often we do that
				yield return new WaitForSeconds(.25f);
			}
		}

		private void Update() {
			// Lol I ain't doing those camera math
			if (Input.GetKey(KeyCode.W)) {
				mainCamera.transform.Translate(new Vector3(-0.1f, 0f, -0.1f), Space.World);
			}
			if (Input.GetKey(KeyCode.S)) {
				mainCamera.transform.Translate(new Vector3(0.1f, 0f, 0.1f), Space.World);
			}
			if (Input.GetKey(KeyCode.A)) {
				mainCamera.transform.Translate(new Vector3(0.1f, 0f, -0.1f), Space.World);
			}
			if (Input.GetKey(KeyCode.D)) {
				mainCamera.transform.Translate(new Vector3(-0.1f, 0f, 0.1f), Space.World);
			}

			// We clicked and we have a character able to do things!
			if (Input.GetMouseButtonDown(0) && _character != null && _character.CurrentState == CharacterState.Waiting) {
				RaycastHit hit;
				Ray ray = mainCamera.ScreenPointToRay(Input.mousePosition);

				if (Physics.Raycast(ray, out hit)) {
					if (hit.transform == null) {
						return;
					}

					TinyExploreArea exploreArea = hit.transform.GetComponent<TinyExploreArea>();
					// We clicked on an explore area
					if (exploreArea != null) {
						//TinyLogger.Log("Yes, this is explore");
						TinyTileMapManager.instance.ActivateExploreArea(exploreArea);
						return;
					}

					TinyDoor door = hit.transform.GetComponent<TinyDoor>();
					// We clicked on a door
					if (door != null) {
						// We are near enough to open the door
						if (Vector3.Distance(_character.transform.position, door.transform.position) < 1.5f) {
							door.OpenDoor();
						}
						return;
					}

					TinyTile tileHit = hit.transform.parent != null ? hit.transform.parent.GetComponent<TinyTile>() : null;
					// We clicked on a tile
					if (tileHit != null) {
						TinySquare destination = TinyTileMapManager.instance.GetSquareAt(hit.point);
						// Was the destination valid?
						if (destination != null && destination.bWalkable) {
							StartCoroutine(WalkThroughPath(TinyTileMapManager.instance.GetPathFinding(TinyTileMapManager.instance.GetSquareAt(_character.transform.position), destination)));
						}

						return;
					}
				}
			}

			// This allows you to get the world cordinates of your click when you use the right mouse button
			if (Input.GetMouseButtonDown(1)) {
				RaycastHit hit;
				Ray ray = mainCamera.ScreenPointToRay(Input.mousePosition);

				if (Physics.Raycast(ray, out hit)) {
					TinyTile tileHit = hit.transform != null ? hit.transform.parent != null ? hit.transform.parent.GetComponent<TinyTile>() : null : null;

					if (tileHit != null) {
						TinyLogger.Log("WorldSpace: " + hit.point);
					}
				}
			}
		}

		/// <summary>
		/// Given an array of Squares it will move the character through them
		/// </summary>
		/// <param name="path">The path to traverse.</param>
		/// <returns></returns>
		IEnumerator WalkThroughPath(TinySquare[] path) {
			if (path.Length == 0) {
				yield break; // If there is no path to walk we give up
			}

			bool bReached = false;
			int index = 1;
			Vector3 targetPos = _character.transform.position;

			_character.SetWalking(true);
			_character.SetState(CharacterState.Moving);

			while (bReached == false) {
				targetPos.x = path[index].x;
				targetPos.z = path[index].y;

				_character.transform.rotation = Quaternion.LookRotation(targetPos - _character.transform.position, Vector3.up);

				while (Vector3.Distance(_character.transform.position, targetPos) > (_character.movementSpeed * Time.deltaTime)) {
					_character.transform.position = Vector3.MoveTowards(_character.transform.position, targetPos, _character.movementSpeed * Time.deltaTime);

					yield return new WaitForEndOfFrame();
				}

				_character.transform.position = targetPos;

				index++;
				if (index >= path.Length) {
					bReached = true;
				}

				yield return new WaitForEndOfFrame();
			}

			_character.SetWalking(false);
			_character.SetState(CharacterState.Waiting);

			TinyTileMapManager.instance.CleanPathFindingMarkers();
			// Here we check if we are at a border so we can explore the new area
			TinyTileMapManager.instance.TestForExploreArea(path[path.Length - 1]);
		}
	}
}