﻿using System;
using System.Collections.Generic;
using UnityEngine;
using Object = UnityEngine.Object;

public static class TinyGameObjectPool {
	private const int DefaultSize = 10;

	//prefab to queue of inactives
	private static readonly Dictionary<GameObject, Queue<GameObject>> Pool = new Dictionary<GameObject, Queue<GameObject>>();
	//instance to prefab
	private static readonly Dictionary<GameObject, GameObject> Active = new Dictionary<GameObject, GameObject>();

	/// <summary>
	/// Spawns some objs from the prefab to allow quick using in runtime
	/// </summary>
	/// <param name="prefab">The prefab.</param>
	/// <param name="count">The count.</param>
	/// <returns></returns>
	public static Queue<GameObject> Warm(GameObject prefab, int count = DefaultSize) {
		Queue<GameObject> queue;
		if (!Pool.TryGetValue(prefab, out queue))
			Pool.Add(prefab, queue = new Queue<GameObject>(count));
		var stored = queue.Count;
		for (int i = 0; i < count - stored; i++) {
			var go = Object.Instantiate(prefab);
			go.SetActive(false);
			queue.Enqueue(go);
		}
		return queue;
	}

	/// <summary>
	/// Gets an instance of the specified prefab.
	/// </summary>
	/// <param name="prefab">The prefab.</param>
	/// <param name="parent">The parent.</param>
	/// <returns></returns>
	public static GameObject Get(GameObject prefab, Transform parent = null) {
		if (!prefab)
			return null; //idiot check

		Queue<GameObject> queue;
		if (!Pool.TryGetValue(prefab, out queue))
			queue = Warm(prefab, DefaultSize);
		GameObject go;
		if (queue.Count == 0 || !(go = queue.Dequeue())) //check if go hasn't been Destroyed
			go = Object.Instantiate(prefab);
		go.transform.SetParent(parent);

		/*ITinyObjectSpawnable tinyObj = go.GetComponent<ITinyObjectSpawnable>();
		if (tinyObj != null) {
			tinyObj.Spawn();
		}*/

		go.SetActive(true);
		Active.Add(go, prefab);
		return go;
	}

	/// <summary>
	/// Gets an instance of the specified prefab.
	/// </summary>
	/// <param name="prefab">The prefab.</param>
	/// <param name="newPosition">The new position.</param>
	/// <param name="parent">The parent.</param>
	/// <returns></returns>
	public static GameObject Get(GameObject prefab, Vector3 newPosition, Transform parent = null) {
		if (!prefab)
			return null; //idiot check

		Queue<GameObject> queue;
		if (!Pool.TryGetValue(prefab, out queue))
			queue = Warm(prefab, DefaultSize);
		GameObject go;
		if (queue.Count == 0 || !(go = queue.Dequeue())) { //check if go hasn't been Destroyed
			go = Object.Instantiate(prefab, newPosition, Quaternion.identity);
		} else {
			go.transform.position = newPosition;
		}
		go.transform.SetParent(parent);

		/*ITinyObjectSpawnable tinyObj = go.GetComponent<ITinyObjectSpawnable>();
		if (tinyObj != null) {
			tinyObj.Spawn();
		}*/

		go.SetActive(true);
		Active.Add(go, prefab);
		return go;
	}

	/// <summary>
	/// Puts the specified gameobject into it's pool.
	/// </summary>
	/// <param name="go">The go.</param>
	public static void Put(GameObject go) {
		if (!go)
			return; //idiot check

		GameObject prefab;
		if (Active.TryGetValue(go, out prefab)) {
			/*ITinyObjectSpawnable tinyObj = go.GetComponent<ITinyObjectSpawnable>();
			if (tinyObj != null) {
				tinyObj.Unspawn();
			}*/

			go.SetActive(false);
			go.transform.SetParent(null);
			go.transform.localPosition = Vector3.zero;
			Active.Remove(go);
			/*foreach (var mono in go.GetComponents<MonoBehaviour>())
				(mono as IDisposable)?.Dispose(); //try dispose*/
			Pool[prefab].Enqueue(go);
		} else
			Object.Destroy(go);
	}
	//replacement for Destroy
	//public static void Put(GameObject go, float delay) => MEC.Timing.CallDelayed(delay, () => Put(go));

	/// <summary>
	/// Destroy all instances in pools.
	/// </summary>
	public static void Clear() {
		foreach (var prefab in Pool.Keys)
			Clear(prefab);
		Pool.Clear();
	}

	/// <summary>
	/// Destroy all instances of specified prefab in pools.
	/// </summary>
	public static void Clear(GameObject prefab) {
		if (!prefab)
			return; //idiot check

		Queue<GameObject> queue;
		if (!Pool.TryGetValue(prefab, out queue))
			return;
		foreach (var go in queue)
			Object.Destroy(go);
		queue.Clear();
	}
}