﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace TinyBravers {

	public class TinyTileMapManager : MonoBehaviour {

		// I'm using odd numbers to have a middle tile
		const int WIDTH = 65;
		const int HEIGHT = 65;

		public static TinyTileMapManager instance;

		/// <summary>
		///  The A* pathfinding system
		/// </summary>
		private TinyPathFinding pathFinding = new TinyPathFinding();

		/// <summary>
		/// The tile map
		/// </summary>
		private TinyTile[,] _tileMap = new TinyTile[WIDTH, HEIGHT];

		private TinyTileEntrance _tileEntrance;

		private int _middleX;
		private int _middleY;

		private List<GameObject> pathFindMarkers = new List<GameObject>(50);
		private TinySquare pathFindInitial;
		private TinySquare pathFindDest;
		private List<TinyExploreArea> _exploreTiles = new List<TinyExploreArea>(10);

		private void Awake() {
			instance = this;

			// Let's cache the middle tile, it's important!
			_middleX = WIDTH / 2;
			_middleY = HEIGHT / 2;
		}

		/// <summary>
		/// Generates the map. Just puts the entrance prefab in the middle of the map and setup the tiles.
		/// </summary>
		public void GenerateMap() {
			_tileEntrance = Instantiate(GameManager.instance.tileDataManager.tileEntrance.gameObject).GetComponent<TinyTileEntrance>();

			_tileMap[_middleX, _middleY] = _tileEntrance.leftTile;
			_tileMap[_middleX + 1, _middleY] = _tileEntrance.rightTile;

			_tileEntrance.leftTile.SetThisTileAt(_middleX, _middleY, GetZeroPositionSquareAtTile(_middleX, _middleY));
			_tileEntrance.rightTile.SetThisTileAt(_middleX + 1, _middleY, GetZeroPositionSquareAtTile(_middleX + 1, _middleY));

			ConnectTwoTiles(_tileEntrance.leftTile, _tileEntrance.rightTile, 2);
		}

		// 0: entrance facing down
		// 1: entrance facing right
		// 2: entrance facing left
		// 3: entrance facing top
		protected void RotateTile(TinyTile tile, byte dir) {
			switch (dir) {
				case 0:
					break;
				case 1:
					tile.Rotate90AntiClockwise();
					tile.Rotate90AntiClockwise();
					tile.Rotate90AntiClockwise();
					tile.gameObject.transform.rotation = Quaternion.Euler(new Vector3(0f, 270f, 0f));
					tile.gameObject.transform.Translate(0f, 0f, 3f, Space.World);
					break;
				case 3:
					tile.Rotate90AntiClockwise();
					tile.Rotate90AntiClockwise();
					tile.gameObject.transform.rotation = Quaternion.Euler(new Vector3(0f, 180f, 0f));
					tile.gameObject.transform.Translate(-3f, 0f, 3f, Space.World);
					break;
				case 2:
					tile.Rotate90AntiClockwise();
					tile.gameObject.transform.rotation = Quaternion.Euler(new Vector3(0f, 90f, 0f));
					tile.gameObject.transform.Translate(-3f, 0f, 0f, Space.World);
					break;
			}
		}

		/// <summary>
		/// Gets the zero position square at tile.
		/// </summary>
		/// <param name="x">The x.</param>
		/// <param name="y">The y.</param>
		/// <returns></returns>
		protected Vector2Int GetZeroPositionSquareAtTile(int x, int y) {
			Vector2Int result = new Vector2Int();

			result.x = (_middleX - x) * 4;
			result.y = (y - _middleY) * 4;

			return result;
		}

		/* 0: tile2 to the top
		 * 1: tile2 to the left
		 * 2: tile2 to the right
		 * 3: tile2 to the bottom
		 */
		protected void ConnectTwoTiles(TinyTile tile1, TinyTile tile2, byte direction) {
			if (tile1 == null || tile2 == null) {
				return;
			}

			tile1.ConnectThisTileTo(tile2, direction);

			switch (direction) {
				case 0:
					tile2.ConnectThisTileTo(tile1, 3);
					break;
				case 1:
					tile2.ConnectThisTileTo(tile1, 2);
					break;
				case 2:
					tile2.ConnectThisTileTo(tile1, 1);
					break;
				case 3:
					tile2.ConnectThisTileTo(tile1, 0);
					break;
			}
		}

		/// <summary>
		/// Gets the initial spawn position.
		/// </summary>
		/// <returns></returns>
		public Vector3 GetInitialSpawnPosition() {
			return _tileEntrance.startingPoints[0] + _tileEntrance.transform.position;
		}

		/// <summary>
		/// Gets the tile grid position given a world position.
		/// </summary>
		/// <param name="worldPos">The world position.</param>
		/// <returns></returns>
		public Vector2Int GetTileGridPosAt(Vector3 worldPos) {
			int iX, iY, x, y;

			int roundedX = Mathf.RoundToInt(worldPos.x);
			int roundedZ = Mathf.RoundToInt(worldPos.z);

			if (roundedX <= 0) {
				iX = (roundedX / 4);
			} else {
				iX = (roundedX / 4);
				if ((roundedX % 4) != 0) {
					iX++;
				}
			}
			if (roundedZ < 0) {
				iY = (roundedZ / 4);
				if ((roundedZ % 4) != 0) {
					iY--;
				}
			} else {
				iY = (roundedZ / 4);
			}

			x = _middleX - iX;
			y = _middleY + iY;

			return new Vector2Int(x, y);
		}

		/// <summary>
		/// Gets the tile at the given world position.
		/// </summary>
		/// <param name="worldPos">The world position.</param>
		/// <returns></returns>
		public TinyTile GetTileAt(Vector3 worldPos) {
			int iX, iY, x, y;

			int roundedX = Mathf.RoundToInt(worldPos.x);
			int roundedZ = Mathf.RoundToInt(worldPos.z);

			if (roundedX <= 0) {
				iX = (roundedX / 4);
			} else {
				iX = (roundedX / 4);
				if ((roundedX % 4) != 0) {
					iX++;
				}
			}
			if (roundedZ < 0) {
				iY = (roundedZ / 4);
				if ((roundedZ % 4) != 0) {
					iY--;
				}
			} else {
				iY = (roundedZ / 4);
			}

			x = _middleX - iX;
			y = _middleY + iY;

			if (_tileMap[x, y] == null) {
				return null;
			}

			return _tileMap[x, y];
		}

		/// <summary>
		/// Gets the square at the given world position.
		/// </summary>
		/// <param name="worldPos">The world position.</param>
		/// <returns></returns>
		public TinySquare GetSquareAt(Vector3 worldPos) {
			int iX, iY, x, y;

			int roundedX = Mathf.RoundToInt(worldPos.x);
			int roundedZ = Mathf.RoundToInt(worldPos.z);

			if (roundedX <= 0) {
				iX = (roundedX / 4);
			} else {
				iX = (roundedX / 4);
				if ((roundedX % 4) != 0) {
					iX++;
				}
			}
			if (roundedZ < 0) {
				iY = (roundedZ / 4);
				if ((roundedZ % 4) != 0) {
					iY--;
				}
			} else {
				iY = (roundedZ / 4);
			}

			x = _middleX - iX;
			y = _middleY + iY;

			if (_tileMap[x, y] == null) {
				return null;
			}

			Vector2Int tilePos = GetZeroPositionSquareAtTile(x, y);

			iX = tilePos.x - roundedX;
			iY = roundedZ - tilePos.y;

			iX = Mathf.Abs(iX);
			iY = Mathf.Abs(iY);

			TinyLogger.Log("tileMap: " + x + ", " + y + " / " + iX + ", " + iY);

			return _tileMap[x, y].squares[iX, iY];
		}

		/// <summary>
		/// Shows the path finding markers.
		/// </summary>
		/// <param name="initial">The initial.</param>
		/// <param name="destination">The destination.</param>
		public void ShowPathFindingMarkers(Vector3 initial, Vector3 destination) {
			ShowPathFindingMarkers(GetSquareAt(initial), GetSquareAt(destination));
		}

		/// <summary>
		/// Shows the path finding markers.
		/// </summary>
		/// <param name="initial">The initial.</param>
		/// <param name="destination">The destination.</param>
		public void ShowPathFindingMarkers(TinySquare initial, TinySquare destination) {
			if (pathFindInitial == initial && pathFindDest == destination) {
				return;
			}

			if (initial == null || destination == null) {
				CleanPathFindingMarkers();
				pathFindInitial = null;
				pathFindDest = null;
				return;
			}

			if (!destination.bWalkable) {
				CleanPathFindingMarkers();
				pathFindInitial = null;
				pathFindDest = null;
				return;
			}

			CleanPathFindingMarkers();

			TinySquare[] path = pathFinding.Pathfind(initial, destination);

			for (int i = 0; i < path.Length; i++) {
				pathFindMarkers.Add(TinyGameObjectPool.Get(GameManager.instance.tileDataManager.pathFindingMarker));
				pathFindMarkers[i].transform.position = new Vector3(path[i].x, pathFindMarkers[i].transform.position.y, path[i].y);
			}

			pathFindInitial = initial;
			pathFindDest = destination;
		}

		/// <summary>
		/// Gets a path from the initial position to the destination.
		/// </summary>
		/// <param name="initial">The initial.</param>
		/// <param name="destination">The destination.</param>
		/// <returns></returns>
		public TinySquare[] GetPathFinding(Vector3 initial, Vector3 destination) {
			return GetPathFinding(GetSquareAt(initial), GetSquareAt(destination));
		}

		/// <summary>
		/// Gets a path from the initial position to the destination.
		/// </summary>
		/// <param name="initial">The initial.</param>
		/// <param name="destination">The destination.</param>
		/// <returns></returns>
		public TinySquare[] GetPathFinding(TinySquare initial, TinySquare destination) {
			return pathFinding.Pathfind(initial, destination);
		}

		/// <summary>
		/// Cleans the path finding markers.
		/// </summary>
		public void CleanPathFindingMarkers() {
			for (int i = 0; i < pathFindMarkers.Count; i++) {
				if (pathFindMarkers[i] != null) {
					TinyGameObjectPool.Put(pathFindMarkers[i]);
				}
			}

			pathFindMarkers.Clear();
		}

		/// <summary>
		/// Tests for explore area at borders.
		/// </summary>
		/// <param name="current">The current position.</param>
		public void TestForExploreArea(TinySquare current) {
			TinyExploreArea exploreTile = null;
			Vector2Int pos = new Vector2Int();

			foreach(TinyExploreArea gObj in _exploreTiles) {
				TinyGameObjectPool.Put(gObj.gameObject);
			}
			_exploreTiles.Clear();

			if (current.topSquare == null) {
				exploreTile = TinyGameObjectPool.Get(GameManager.instance.tileDataManager.exploreTile).GetComponent<TinyExploreArea>();
				pos = GetTileGridPosAt(new Vector3(current.x, 0f, current.y - 1));
				exploreTile.direction = 0;

				exploreTile.x = pos.x;
				exploreTile.y = pos.y;

				float x = -1.5f + ((_middleX - pos.x) * 4);
				float y = 1.5f + ((pos.y - _middleY) * 4);

				exploreTile.transform.position = new Vector3(x, -0.5f, y);

				_exploreTiles.Add(exploreTile);
			}
			if (current.leftSquare == null) {
				exploreTile = TinyGameObjectPool.Get(GameManager.instance.tileDataManager.exploreTile).GetComponent<TinyExploreArea>();
				pos = GetTileGridPosAt(new Vector3(current.x + 1, 0f, current.y));
				exploreTile.direction = 1;

				exploreTile.x = pos.x;
				exploreTile.y = pos.y;

				float x = -1.5f + ((_middleX - pos.x) * 4);
				float y = 1.5f + ((pos.y - _middleY) * 4);

				exploreTile.transform.position = new Vector3(x, -0.5f, y);

				_exploreTiles.Add(exploreTile);
			}
			if (current.rightSquare == null) {
				exploreTile = TinyGameObjectPool.Get(GameManager.instance.tileDataManager.exploreTile).GetComponent<TinyExploreArea>();
				pos = GetTileGridPosAt(new Vector3(current.x - 1, 0f, current.y));
				exploreTile.direction = 2;

				exploreTile.x = pos.x;
				exploreTile.y = pos.y;

				float x = -1.5f + ((_middleX - pos.x) * 4);
				float y = 1.5f + ((pos.y - _middleY) * 4);

				exploreTile.transform.position = new Vector3(x, -0.5f, y);

				_exploreTiles.Add(exploreTile);
			}
			if (current.bottomSquare == null) {
				exploreTile = TinyGameObjectPool.Get(GameManager.instance.tileDataManager.exploreTile).GetComponent<TinyExploreArea>();
				pos = GetTileGridPosAt(new Vector3(current.x, 0f, current.y + 1));
				exploreTile.direction = 3;

				exploreTile.x = pos.x;
				exploreTile.y = pos.y;

				float x = -1.5f + ((_middleX - pos.x) * 4);
				float y = 1.5f + ((pos.y - _middleY) * 4);

				exploreTile.transform.position = new Vector3(x, -0.5f, y);

				_exploreTiles.Add(exploreTile);
			}
		}

		/// <summary>
		/// Clears the explore tile at.
		/// </summary>
		/// <param name="x">The x.</param>
		/// <param name="y">The y.</param>
		public void ClearExploreTileAt(int x, int y) {
			foreach (TinyExploreArea exp in _exploreTiles) {
				if (exp.x == x && exp.y == y) {
					TinyGameObjectPool.Put(exp.gameObject);
					_exploreTiles.Remove(exp);
					break;
				}
			}
		}

		/// <summary>
		/// Explores the are and generates a new tile.
		/// </summary>
		/// <param name="explore">The explore.</param>
		public void ActivateExploreArea(TinyExploreArea explore) {
			GameObject randomTile = GameManager.instance.tileDataManager.tiles[Random.Range(0, GameManager.instance.tileDataManager.tiles.Length)];
			//GameObject randomTile = GameManager.instance.tileDataManager.tiles[1];

			TinyTile newTile = Instantiate(randomTile).GetComponent<TinyTile>();

			ClearExploreTileAt(explore.x, explore.y);
			newTile.transform.position = new Vector3((_middleX - explore.x) * 4, -0.5f, (explore.y - _middleY) * 4);

			RotateTile(newTile, explore.direction);
			_tileMap[explore.x, explore.y] = newTile;

			newTile.SetThisTileAt(explore.x, explore.y, GetZeroPositionSquareAtTile(explore.x, explore.y));

			ConnectTwoTiles(newTile, _tileMap[explore.x, explore.y - 1], 0); //top
			ConnectTwoTiles(newTile, _tileMap[explore.x - 1, explore.y], 1); //left
			ConnectTwoTiles(newTile, _tileMap[explore.x + 1, explore.y], 2); //right
			ConnectTwoTiles(newTile, _tileMap[explore.x, explore.y + 1], 3); //down
		}
	}
}