﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace TinyBravers {

	public enum SquareType {
		Empty,
		Wall,
		Door,
	}

	/// <summary>
	/// A tile is a group of squares.
	/// </summary>
	/// <seealso cref="UnityEngine.MonoBehaviour" />
	public class TinyTile : MonoBehaviour {

		[SerializeField]
		private SquareType[] _squareTypes = new SquareType[16];

		[SerializeField]
		private GameObject[] _obstacles = new GameObject[16];

		[NonSerialized]
		public TinySquare[,] squares = new TinySquare[4,4];

		[NonSerialized]
		public Vector2Int tilePositionInGrid = new Vector2Int();

		private SquareType GetSquareType(int x, int y) {
			return _squareTypes[y * 4 + x];
		}

		private GameObject GetObstacle(int x, int y) {
			return _obstacles[y * 4 + x];
		}

		private int ConvertIndex(int x, int y) {
			return y * 4 + x;
		}

		public SquareType GetSquareTypeAt(int x, int y) {
			return GetSquareType(x, y);
		}

		public TinySquare SafeGetSquare(int x, int y) {
			if (x < 0 || x >= squares.GetLength(0)) {
				return null;
			}

			if (y < 0 || y >= squares.GetLength(1)) {
				return null;
			}

			return squares[x, y];
		}

		private void Awake() {
			for (int x = 0; x < squares.GetLength(0); x++) {
				for (int y = 0; y < squares.GetLength(1); y++) {
					squares[x, y] = new TinySquare(GetSquareTypeAt(x, y));
				}
			}
		}

		public void Rotate90AntiClockwise() {
			RotateMatrix.rotate90(squares, 4);
		}

		public void SetObstacleAt(SquareType type, int x, int y) {
			if (GetObstacle(x, y) != null) {
				DestroyImmediate(_obstacles[ConvertIndex(x, y)]);
			}

			switch (type) {
				case SquareType.Wall:
					if (x == 0 && y == 0) {
						if ((GetSquareTypeAt(0, 1) == SquareType.Wall || GetSquareTypeAt(0, 1) == SquareType.Door) && (GetSquareTypeAt(1, 0) == SquareType.Wall || GetSquareTypeAt(1, 0) == SquareType.Door)) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wallCorner, new Vector3(-x + 0.5f, 0.5f, y - 0.5f) + transform.position, Quaternion.Euler(0f, 180f, 0f), transform);
						} else if (GetSquareTypeAt(1, 0) == SquareType.Wall || GetSquareTypeAt(1, 0) == SquareType.Door) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 270f, 0f), transform);
						} else {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 180f, 0f), transform);
						}
					} else if (x == 3 && y == 0) {
						if ((GetSquareTypeAt(3, 1) == SquareType.Wall || GetSquareTypeAt(3, 1) == SquareType.Door) && (GetSquareTypeAt(2, 0) == SquareType.Wall || GetSquareTypeAt(2, 0) == SquareType.Door)) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wallCorner, new Vector3(-x - 0.5f, 0.5f, y - 0.5f) + transform.position, Quaternion.Euler(0f, 270f, 0f), transform);
						} else if (GetSquareTypeAt(2, 0) == SquareType.Wall || GetSquareTypeAt(2, 0) == SquareType.Door) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 270f, 0f), transform);
						} else {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 0f, 0f), transform);
						}
					} else {
						if (x == 0) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 180f, 0f), transform);
						}
						if (x == 1 || x == 2) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 270f, 0f), transform);
						}
						if (x == 3) {
							_obstacles[ConvertIndex(x, y)] = Instantiate(GameManager.GetFromEditor.tileDataManager.wall, new Vector3(-x, 0.5f, y) + transform.position, Quaternion.Euler(0f, 0f, 0f), transform);
						}
					}

					_squareTypes[ConvertIndex(x, y)] = type;
					break;
				case SquareType.Door:
					RemoveDoors();

					if (x == 0) {
						SetDoorAt(2);
					} else if (x == 1 || x == 2) {
						SetDoorAt(0);
					} else {
						SetDoorAt(1);
					}
					break;
				case SquareType.Empty:
					_squareTypes[ConvertIndex(x, y)] = type;
					break;
			}
		}

		/* 0: top
		 * 1: right
		 * 2: left */
		protected void SetDoorAt(byte direction) {
			switch (direction) {
				case 0:
					SetObstacleAt(SquareType.Empty, 1, 0);
					SetObstacleAt(SquareType.Empty, 2, 0);
					_squareTypes[ConvertIndex(1, 0)] = SquareType.Door;
					_squareTypes[ConvertIndex(2, 0)] = SquareType.Door;

					_obstacles[ConvertIndex(1, 0)] = Instantiate(GameManager.GetFromEditor.tileDataManager.door, new Vector3(-1.5f, 0.5f, 0f) + transform.position, Quaternion.Euler(0f, 0f, 0f), transform);
					_obstacles[ConvertIndex(2, 0)] = _obstacles[ConvertIndex(1, 0)];
					break;
				case 1:
					SetObstacleAt(SquareType.Empty, 3, 1);
					SetObstacleAt(SquareType.Empty, 3, 2);
					_squareTypes[ConvertIndex(3, 1)] = SquareType.Door;
					_squareTypes[ConvertIndex(3, 2)] = SquareType.Door;

					_obstacles[ConvertIndex(3, 1)] = Instantiate(GameManager.GetFromEditor.tileDataManager.door, new Vector3(-3f, 0.5f, 1.5f) + transform.position, Quaternion.Euler(0f, 90f, 0f), transform);
					_obstacles[ConvertIndex(3, 2)] = _obstacles[ConvertIndex(3, 1)];
					break;
				case 2:
					SetObstacleAt(SquareType.Empty, 0, 1);
					SetObstacleAt(SquareType.Empty, 0, 2);
					_squareTypes[ConvertIndex(0, 1)] = SquareType.Door;
					_squareTypes[ConvertIndex(0, 2)] = SquareType.Door;

					_obstacles[ConvertIndex(0, 1)] = Instantiate(GameManager.GetFromEditor.tileDataManager.door, new Vector3(0f, 0.5f, 1.5f) + transform.position, Quaternion.Euler(0f, 270f, 0f), transform);
					_obstacles[ConvertIndex(0, 1)] = _obstacles[ConvertIndex(0, 1)];
					break;
			}
		}

		public void RemoveDoors() {
			for (int x = 0; x < 4; x++) {
				for (int y = 0; y < 4; y++) {
					if (_squareTypes[ConvertIndex(x, y)] == SquareType.Door || _squareTypes[ConvertIndex(x, y)] == SquareType.Door) {
						_squareTypes[ConvertIndex(x, y)] = SquareType.Empty;
						DestroyImmediate(_obstacles[ConvertIndex(x, y)]);
					}
				}
			}
		}

		public void FixPositions() {
			transform.position = Vector3.zero;

			foreach (Transform child in transform) {
				child.Translate(new Vector3(-1.5f, 0f, 1.5f) + transform.position);
			}
		}

		public void SetThisTileAt(int newX, int newY, Vector2Int zeroSquarePosition) {
			tilePositionInGrid.x = newX;
			tilePositionInGrid.y = newY;

			for (int x = 0; x < 4; x++) {
				for (int y = 0; y < 4; y++) {
					squares[x, y].x = zeroSquarePosition.x - x;
					squares[x, y].y = zeroSquarePosition.y + y;

					squares[x, y].topSquare = SafeGetSquare(x, y - 1);
					squares[x, y].leftSquare = SafeGetSquare(x - 1, y);
					squares[x, y].rightSquare = SafeGetSquare(x + 1, y);
					squares[x, y].bottomSquare = SafeGetSquare(x, y + 1);
				}
			}
		}

		/* 0: tile2 to the top
		 * 1: tile2 to the left
		 * 2: tile2 to the right
		 * 3: tile2 to the bottom
		 */
		public void ConnectThisTileTo(TinyTile tile2, byte direction) {
			switch (direction) {
				case 0:
					for (int x = 0; x < 4; x++) {
						squares[x, 0].topSquare = tile2.squares[x, 3];
					}
					break;
				case 1:
					for (int y = 0; y < 4; y++) {
						squares[0, y].leftSquare = tile2.squares[3, y];
					}
					break;
				case 2:
					for (int y = 0; y < 4; y++) {
						squares[3, y].rightSquare = tile2.squares[0, y];
					}
					break;
				case 3:
					for (int x = 0; x < 4; x++) {
						squares[x, 3].bottomSquare = tile2.squares[x, 0];
					}
					break;
			}
		}
	}
}